import click
import webbrowser
import yaml
import keyring

from rich.console import Console

from bytewax.client import Client
from bytewax.errors import Error
from bytewax.client.utils import (
    get_device_code_url,
    begin_polling_loop,
    store_auth_info,
    load_api_token,
    delete_token,
)
from bytewax.gateway import gateway as gtw

console = Console()

_global_cli_options = [
    click.option(
        "--verbose",
        "-v",
        is_flag=True,
        default=False,
        help="Verbose output",
    )
]


def global_cli_options(func):
    for option in reversed(_global_cli_options):
        func = option(func)
    return func


@click.group()
def cli():
    pass


######### API METHODS #############


@click.command()
@global_cli_options
def list_swarms(verbose):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.list_swarms())
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_name", required=True)
def delete_swarm(verbose, swarm_name):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.delete_swarm(swarm_name))
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_definition_file", required=True, type=click.File("r"))
def deploy(verbose, swarm_definition_file):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        try:
            swarm_definition = yaml.safe_load(swarm_definition_file)
        except yaml.YAMLError as err:
            console.print(f"Error parsing swarm definition file: {err}")
            return

        client.deploy(swarm_definition)
        print(f"Deployed swarm.")
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_name", required=True)
def get_swarm(verbose, swarm_name):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.get_swarm(swarm_name))
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_name", required=True)
def get_bees(verbose, swarm_name):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.get_bees(swarm_name))
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_name", required=True)
@click.argument("bee_name", required=True)
def get_bee(verbose, swarm_name, bee_name):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.get_bee(swarm_name, bee_name))
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("swarm_name", required=True)
@click.argument("bee_name", required=True)
def get_bee_logs(verbose, swarm_name, bee_name):
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.get_bee_logs(swarm_name, bee_name))
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
def login(verbose):
    console.print("Welcome to bytewax :bee:")
    auth_response = get_device_code_url(
        Client.required_scopes,
    )

    console.print(
        f"Your device authentication code is: [bold]{auth_response['user_code']}"
    )

    if click.confirm(
        "Press Enter to open a web browser to login or ^C to quit",
        prompt_suffix="",
        show_default=False,
        default=True,
    ):
        try:
            with console.status("Waiting for login to complete in browser"):
                console.print(
                    f'Opening login page {auth_response["verification_uri_complete"]}'
                )
                webbrowser.open(auth_response["verification_uri_complete"])
                # Start polling for a successful login from auth0
                token_response = begin_polling_loop(auth_response)
                try:
                    store_auth_info(token_response)
                except keyring.errors.NoKeyringError:
                    console.print("No keyring in the system - refresh token disabled")
                console.print("Successfully logged in.")
        except Error as err:
            console.print(f"{err}")


@click.command()
@global_cli_options
def logout(verbose):
    console.print(f"Logging out of bytewax.")
    delete_token()


######### GATEWAY METHODS #############
@click.command()
@global_cli_options
@click.option("--gateway", "-g", required=True, help="Gateway name to publish to")
def request(verbose, gateway):
    """
    Make request to gateway. Usage `echo mypayload | waxctl request --gateway gatewayname`
    """
    payload = {"stdin": click.get_text_stream("stdin").read()}
    gtw_obj = gtw.Gateway()
    response = gtw_obj.request(payload, gateway=gateway)
    console.print(f"{response}")


######### SECRETS METHODS #############
@click.command()
@global_cli_options
def list_secrets(verbose):
    """
    List secrets
    """
    try:
        token = load_api_token()
        client = Client(token, verbose)
        console.print(client.list_secrets())
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("name", required=True)
@click.argument("content", required=True)
def set_secret(verbose, name, content):
    """
    Create or update secret
    """
    try:
        token = load_api_token()
        client = Client(token, verbose)
        client.set_secret(name, content)
        console.print("Secret set")
    except Error as err:
        console.print(f"{err}")


@click.command()
@global_cli_options
@click.argument("name", required=True)
def delete_secret(verbose, name):
    """
    Create or update secret
    """
    try:
        token = load_api_token()
        client = Client(token, verbose)
        client.delete_secret(name)
        console.print("Secret deleted")
    except Error as err:
        console.print(f"{err}")


cli.add_command(list_swarms)
cli.add_command(deploy)
cli.add_command(delete_swarm)
cli.add_command(get_swarm)
cli.add_command(get_bees)
cli.add_command(get_bee)
cli.add_command(get_bee_logs)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(request)
cli.add_command(list_secrets)
cli.add_command(set_secret)
cli.add_command(delete_secret)
