from __future__ import annotations

from typing import List, Dict
import os
import logging

from requests import Session

from bytewax.errors import (
    Error,
    SwarmNotFound,
    SwarmCreationError,
    SwarmDeletionError,
    ResourceNotFound,
)
from .utils import timeout_handler

API_HOST = os.getenv("BYTEWAX_API_HOST", "https://api.bytewax.io")

logger = logging.getLogger(__name__)


class Client(object):
    required_scopes = [
        "openid",
        "offline_access",  # to get a refresh token
        "profile",
        "email",
    ]

    def __init__(self, token, verbose=False):
        self.session = Session()
        self.session.headers.update({"Authorization": f"Bearer {token}"})
        self.setup_logging(verbose)

    def setup_logging(self, verbose):
        if verbose:
            logging_level = logging.DEBUG
        else:
            logging_level = logging.INFO

        logging.getLogger("urllib3").setLevel(logging_level)
        logging.getLogger().setLevel(logging_level)

    def list_swarms(self) -> List[Dict]:
        """List all swarms

        Returns:
            obj:`list`:
                A list of swarms
        """
        with timeout_handler():
            response = self.session.get(f"{API_HOST}/swarms", timeout=10)

        if response.ok:
            return response.json()
        elif response.status_code == 404:
            raise Error(f"No swarms found.")
        else:
            raise Error(f"Error listing swarms: {response.json()}")

    def deploy(self, swarm_definition: Dict):
        """Deploy new swarm or update existing

        Args:
            swarm_definition (:obj:`dict`):
                A dictionary describing the swarm.
        """
        with timeout_handler():
            response = self.session.post(
                f"{API_HOST}/swarms",
                timeout=10,
                json=swarm_definition,
            )

        if response.ok:
            return response.json()
        else:
            raise SwarmCreationError(
                f"Error creating swarm: {response.json()}"
            )

    def delete_swarm(self, swarm_name):
        """Delete ``swarm_name``

        Args:
            swarm_name (:obj:`string`):
                The name of the swarm to delete.
        """
        with timeout_handler():
            response = self.session.delete(
                f"{API_HOST}/swarms/{swarm_name}", timeout=10
            )

        if response.ok:
            return f"Swarm {swarm_name} was deleted."
        else:
            raise SwarmDeletionError(f"Error deleting swarm: {response.json()}")

    def get_swarm(self, swarm_name: str) -> Dict:
        """Get information about ``swarm_name``

        Args:
            swarm_name (:obj:`string`):
                The name of the swarm to query.
        Returns:
            :obj:`dict`:
                A dict describing the swarm and its current state.
        """
        with timeout_handler():
            response = self.session.get(f"{API_HOST}/swarms/{swarm_name}", timeout=10)

        if response.ok:
            return response.json()
        elif response.status_code == 404:
            raise SwarmNotFound(f"Swarm {swarm_name} not found.")
        else:
            raise Error(f"Error when getting swarm: {response.json()}")

    def get_bees(self, swarm_name: str) -> Dict:
        """Get information about bees in ``swarm_name``

        Args:
            swarm_name (:obj:`string`):
                The name of the swarm to query.
        Returns:
            :obj:`dict`:
                A dict describing the bees contained in the current swarm.
        """
        with timeout_handler():
            response = self.session.get(
                f"{API_HOST}/swarms/{swarm_name}/bees", timeout=10
            )

        if response.ok:
            return response.json()
        elif response.status_code == 404:
            raise SwarmNotFound(f"Swarm {swarm_name} not found.")
        else:
            raise Error(f"Error when getting bees from {swarm_name}: {response.json()}")

    def get_bee(self, swarm_name, bee_name):
        """Get information about the bee named ``bee_name`` in ``swarm_name``

        Args:
            swarm_name (:obj:`string`):
                The name of the swarm to query.
            bee_name (:obj:`string`):
                The name of the bee to query.
        Returns:
            :obj:`dict`:
                A dict describing the bee.
        """
        with timeout_handler():
            response = self.session.get(
                f"{API_HOST}/swarms/{swarm_name}/bees/{bee_name}",
                timeout=10,
            )

        if response.ok:
            return response.json()
        elif response.status_code == 404:
            raise ResourceNotFound(f"Not found: {response.json()['message']}")
        else:
            raise Error(f"Error when getting bee from {swarm_name}: {response.json()}")

    def get_bee_logs(self, swarm_name, bee_name):
        """Fetch the logs for the bee named ``bee_name`` in ``swarm_name``

        Args:
            swarm_name (:obj:`string`):
                The name of the swarm to query.
            bee_name (:obj:`string`):
                The name of the bee to query.
        Returns:
            :obj:`dict`:
                A dict describing the bee.
        """
        with timeout_handler():
            response = self.session.get(
                f"{API_HOST}/swarms/{swarm_name}/bees/{bee_name}/logs",
                timeout=10,
            )

        if response.ok:
            return response.json()
        elif response.status_code == 404:
            raise ResourceNotFound(f"Not found: {response.json()['message']}")
        else:
            raise Error(
                f"Error when getting bee logs from bee {bee_name}: {response.json()}"
            )

    def list_secrets(self):
        """Fetch list of secrets

        Returns:
            :obj:`list`:
                A list of secrets.
        """
        with timeout_handler():
            response = self.session.get(
                f"{API_HOST}/secrets",
                timeout=10,
            )
        if response.ok:
            return response.json()
        else:
            raise Error(f"Error when listing secrets: {response.json()}")

    def set_secret(self, name, content):
        """Creates or updates secret

        Args:
            name (:obj:`string`):
                The name of secret
            content (:obj:`string`):
                Contents of secret
        """
        with timeout_handler():
            response = self.session.post(
                f"{API_HOST}/secrets",
                timeout=10,
                json={"name": name, "content": content},
            )
        if response.ok:
            return response.json()
        else:
            raise Error(f"Error when setting secrets: {response.json()}")

    def delete_secret(self, name):
        """Deletes secret

        Args:
            name (:obj:`string`):
                The name of secret
        """
        with timeout_handler():
            response = self.session.delete(
                f"{API_HOST}/secrets/{name}",
                timeout=10,
            )
        if response.ok:
            return response.json()
        else:
            raise Error(f"Error when deleting secrets: {response.json()}")
