import time
import os
import requests
import configparser
import keyring
import jwt

from contextlib import contextmanager
from pathlib import Path

from jwt import PyJWKClient

from cachetools import cached, TTLCache

from bytewax.errors import Unavailable, Error


BYTEWAX_CONF_FILE = os.getenv("BYTEWAX_CONF_FILE", f"{Path.home()}/.bytewax")


@contextmanager
def timeout_handler():
    try:
        yield
    except requests.ConnectionError as e:
        raise Unavailable(
            "Connection error when connecting to the bytewax API endpoint"
        ) from e
    except TimeoutError as e:
        raise Unavailable("Timeout when connecting to the bytewax API endpoint") from e


def get_device_code_url(
    scopes, domain="bytewax.us.auth0.com", client_id="MYwiRow5MmhHhCTKlen7dVJyGmB8LZDZ"
):
    """Build an auth0 device code URL for the CLI to open in a browser"""

    payload = {
        "client_id": client_id,
        "scope": " ".join(scopes),  # scopes have to be joined by a space
        "audience": "https://api.bytewax.net",
    }
    response = requests.post(f"https://{domain}/oauth/device/code", data=payload)

    if response.ok:
        return response.json()
    else:
        response.raise_for_status()


def begin_polling_loop(
    auth_response,
    domain="bytewax.us.auth0.com",
    client_id="MYwiRow5MmhHhCTKlen7dVJyGmB8LZDZ",
):
    """Poll auth0 for successful login"""

    payload = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": auth_response["device_code"],
        "client_id": client_id,
    }

    while True:
        response = requests.post(f"https://{domain}/oauth/token", payload)

        if response.ok:
            return response.json()
        # TODO: Handle denied flow
        else:
            time.sleep(auth_response["interval"])


def store_auth_info(token_response):
    """Store the JWT token and refresh token

    Generated JWT tokens and the assosciated namespace are stored in the
    :func:`~Path.home` directory in a file named ``.bytewax``.

    The refresh tokens are stored in the system keyring.
    """
    config = configparser.ConfigParser()
    config.read(BYTEWAX_CONF_FILE)
    token = token_response["access_token"]

    signing_key = get_signing_key(token)
    # Validate the token before storing it

    token_payload = jwt.decode(
        token,
        signing_key.key,
        audience="https://api.bytewax.net",
        algorithms=["RS256"],
    )

    config["DEFAULT"] = {
        "access_token": token,
        "namespace": token_payload["https://bytewax.net/app_metadata"]["namespace"],
    }

    with open(BYTEWAX_CONF_FILE, "w") as configfile:
        config.write(configfile)

    # Store the refresh token in the system keyring
    keyring.set_password(
        "system", "bytewax_refresh_token", token_response["refresh_token"]
    )


@cached(cache=TTLCache(maxsize=1, ttl=600))
def get_signing_key(token):
    """Return the auth0 signing key for our tenant"""
    jwks_client = PyJWKClient("https://bytewax.us.auth0.com/.well-known/jwks.json")
    return jwks_client.get_signing_key_from_jwt(token)


def get_gateway_token(
    client_id, client_secret, audience, domain="bytewax.us.auth0.com"
):
    """Create a M2M token from Auth0, and save it in .bytewax"""
    payload = {
        "grant_type": "client_credentials",
        "client_id": client_id,
        "client_secret": client_secret,
        "audience": audience,
    }
    response = requests.post(f"https://{domain}/oauth/token", json=payload)
    if response.ok:
        token_response = response.json()
        token = token_response["access_token"]
        config = configparser.ConfigParser()
        config.read(BYTEWAX_CONF_FILE)
        config["GATEWAY"] = {"access_token": token, "audience": audience}
        with open(BYTEWAX_CONF_FILE, "w") as configfile:
            config.write(configfile)
        return token
    else:
        response.raise_for_status()
        return Error("Error when fetching token from Auth0.")


def load_gateway_token(
    client_id: str, client_secret: str, audience="https://gateway.bytewax.io"
):
    config = configparser.ConfigParser()
    config.read(BYTEWAX_CONF_FILE)
    try:
        token = config["GATEWAY"]["access_token"]
    except KeyError:
        token = get_gateway_token(client_id, client_secret, audience)

    signing_key = get_signing_key(token)
    try:
        jwt.decode(
            token,
            signing_key.key,
            audience=audience,
            algorithms=["RS256"],
        )

        return token
    except jwt.ExpiredSignatureError:
        return get_gateway_token(client_id, client_secret, audience)


def load_api_token(
    domain="bytewax.us.auth0.com", client_id="MYwiRow5MmhHhCTKlen7dVJyGmB8LZDZ"
):
    """Load the JWT token for authentication

    If the token has expired, call :func:`~refresh_token` to generate
    a new one.
    """
    config = configparser.ConfigParser()
    config.read(BYTEWAX_CONF_FILE)
    try:
        token = config["DEFAULT"]["access_token"]
    except KeyError:
        raise (
            Error(
                "Could not load bytewax authentication information. Have you logged in with `waxctl login`?"
            )
        )

    # Validate and decode the token.
    # if the token is expired, it will throw ExpiredSignatureError
    signing_key = get_signing_key(token)
    try:
        jwt.decode(
            token,
            signing_key.key,
            audience="https://api.bytewax.net",
            algorithms=["RS256"],
        )

        return token
    except jwt.ExpiredSignatureError:
        return refresh_token(domain, client_id)


def refresh_token(
    domain="bytewax.us.auth0.com", client_id="MYwiRow5MmhHhCTKlen7dVJyGmB8LZDZ"
):
    """Regenerate a JWT token

    Load the refresh token from the system keyring, and use it to
    generate a new JWT token.
    """
    refresh_token = keyring.get_password("system", "bytewax_refresh_token")

    payload = {
        "grant_type": "refresh_token",
        "client_id": client_id,
        "refresh_token": refresh_token,
    }

    response = requests.post(f"https://{domain}/oauth/token", data=payload)
    if response.ok:
        token_response = response.json()
        store_auth_info(token_response)

        # Return the access token here for use in the current request
        return token_response["access_token"]
    else:
        response.raise_for_status()


def delete_token():
    """Delete the current JWT token from the .bytewax file"""
    try:
        keyring.delete_password("system", "bytewax_refresh_token")
    except keyring.errors.PasswordDeleteError:
        pass

    try:
        credential_file = BYTEWAX_CONF_FILE
        open(credential_file)
        config = configparser.ConfigParser()
        config.read(credential_file)
        config["DEFAULT"]["access_token"] = ""
    except IOError:
        pass
