from __future__ import annotations

import sys
import os
import logging
import grpc
import json
import uuid
import structlog

from typing import Dict, Union
from bytewax.client.utils import load_api_token, load_gateway_token

from bytewax.errors import GatewayTimeoutError, Error

from .proto import gateway_pb2, gateway_pb2_grpc

log = structlog.get_logger()


class Gateway(object):
    def __init__(self, client_id: str = None, client_secret: str = None):
        # TODO: This is a hack to allow users to use their personal tokens
        # To call the bytewax Gateway using the token we generate for
        # api.bytewax.net
        # This only works because we don't (currently) validate the audience
        # param in the JWT token on the gateway. We should do that.
        self.client_id = client_id
        self.client_secret = client_secret

        self.swarm_api_address = os.getenv("BYTEWAX_GATEWAY", "gateway.bytewax.io:443")
        if os.getenv("BYTEWAX_GATEWAY_INSECURE"):
            channel = grpc.insecure_channel(self.swarm_api_address)
        else:
            credentials = grpc.ssl_channel_credentials()
            channel = grpc.secure_channel(self.swarm_api_address, credentials)
        # Create the gRPC service stub
        self.stub = gateway_pb2_grpc.GatewayStub(channel)

    @property
    def token(self):
        if self.client_id == None and self.client_secret == None:
            return load_api_token()
        else:
            return load_gateway_token(self.client_id, self.client_secret)

    def submit(self, payload: Union[Dict, str], gateway: str, timeout: float = 30.0):
        """Send ``payload`` to the Gateway via gRPC.

        This method sends a ``payload`` for processing to the gateway
        but does not expect a response.

        Args:
            payload(:obj:`Union[Dict, str]`):
                A dict or string to be processed by the Swarm.
            gateway(:obj:`str`):
                The name of the gateway to send the payload to.
            timeout(:obj:`float`):
                The gRPC deadline(timeout) in seconds.
        """
        request_id = str(uuid.uuid4())

        message = gateway_pb2.GatewayMessage()
        # TODO include pickling if needed, and context
        message.payload = json.dumps(payload)
        message.token = self.token
        message.name = gateway
        message.request_id = request_id

        log.info(f"Submitting request", request_id=request_id)
        return self.stub.Submit(message, timeout=timeout)

    def request(
        self, payload: Union[Dict, str], gateway: str, timeout: float = 30.0
    ) -> gateway_pb2.GatewayResponse:
        """Send ``payload`` to the Gateway via gRPC.

        This method sends a ``payload`` for processing to the gateway
        that _should_ expect a response.

        Args:
            payload(:obj:`Union[Dict, str]`):
                A dict or string to be processed by the Swarm.
            gateway(:obj:`str`):
                The name of the gateway to send the payload to.
            timeout(:obj:`float`):
                The gRPC deadline(timeout) in seconds.
        """
        request_id = str(uuid.uuid4())

        message = gateway_pb2.GatewayMessage()
        # TODO include pickling if needed, and context
        message.payload = json.dumps(payload)
        message.token = self.token
        message.name = gateway
        message.request_id = request_id

        log.info(f"Sending request", request_id=request_id)
        try:
            return self.stub.Request(message, timeout=timeout)
        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                raise GatewayTimeoutError(
                    f"Timed out waiting for the Gateway to respond to request_id {request_id}. Check to make sure that the gateway param `{gateway}` matches the gateway input name defined in your swarm.yaml, and that your Swarm is healthy."
                )
            else:
                raise Error(
                    f"Received unknown RPC error: code={rpc_error.code()} message={rpc_error.details()}"
                )
