from __future__ import annotations

import sys
import logging
import base64
import pickle
import time
import traceback
import signal
import structlog

from collections.abc import Callable
from typing import Dict, Union, Any

from dataclasses import dataclass

from bytewax.swarm.metrics import *

import grpc
import json

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bytewax.swarm import Swarm, DevSwarm

from bytewax.swarm.proto import swarm_pb2, swarm_pb2_grpc

log = structlog.get_logger()

from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(repr=True)
class InputType:
    name: str
    queue_name: str
    swarm_name: str
    user: str
    swarm_api_address: str
    token: str
    swarm_version: str


@dataclass(repr=True)
class BeeInput(InputType):
    def watch(
        self,
        swarm: Union[Swarm, DevSwarm],
        fn: Callable[[Union[Swarm, DevSwarm], Union[str, Dict], Dict[Any, Any]], Any],
        watcher,
    ):
        """Connect to the swarm-api and consume messages

        Constructs a gRPC `Bee` to take input from the swarm-api
        process and calls `fn` with the swarm, payload and context.

        Args:
            swarm (Swarm, DevSwarm): An instance of the swarm in use.
            fn: (function): The function to call for each message.
        """
        # Create a gRPC channel to the swarm-api
        channel = grpc.insecure_channel(self.swarm_api_address)

        # Construct the gRPC Bee
        bee = swarm_pb2.Bee(
            token=self.token,
            swarm=self.swarm_name,
            name=self.name,
            swarm_version=self.swarm_version,
            queue_name=self.queue_name,
        )

        # Create the gRPC service stub
        self.stub = swarm_pb2_grpc.SwarmServiceStub(channel)

        log.debug(f"Start watching bee {bee}")
        for inp in self.stub.Watch(bee):
            log.debug(f"Got message {inp}")
            # Add the request_id and timeout to the context
            inp.context["request_id"] = inp.request_id
            inp.context["_bw_request_timeout"] = inp.timeout.ToJsonString()

            start_time = time.time()
            if inp.context.get("_bw_message_type") == "pickle":
                pkl = base64.b16decode(inp.payload)
                payload = pickle.loads(pkl)
            else:
                payload = json.loads(inp.payload)
            try:
                fn(swarm, payload, inp.context)
                messages_received.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name, 0
                ).inc()
            except Exception as e:
                messages_received.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name, 1
                ).inc()
                (_, _, tb) = sys.exc_info()
                error_message = traceback.format_exception(type(e), e, tb)
                swarm.respond(
                    f"Exception when handling payload: {error_message}", inp.context
                )
                raise e
            finally:
                end_time = time.time() - start_time
                bee_processing_time.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name
                ).observe(end_time)


@dataclass(repr=True)
class GatewayInput(InputType):
    def watch(
        self,
        swarm: Union[Swarm, DevSwarm],
        fn: Callable[[Union[Swarm, DevSwarm], Union[str, Dict], Dict[Any, Any]], Any],
        watcher,
    ):
        """Connect to the swarm-api and consume messages

        Constructs a gRPC `Bee` to take input from the swarm-api
        process and calls `fn` with the swarm, payload and context.

        Args:
            swarm (Swarm, DevSwarm): An instance of the swarm in use.
            fn: (function): The function to call for each message.
        """
        # Create a gRPC channel to the swarm-api
        channel = grpc.insecure_channel(self.swarm_api_address)

        gateway = swarm_pb2.GatewayName(
            name=self.name,
            token=self.token,
            queue_name=self.queue_name,
        )

        # Create the gRPC service stub
        self.stub = swarm_pb2_grpc.SwarmServiceStub(channel)

        log.debug(f"Start watching gateway: {gateway}")
        stream = self.stub.WatchGateway(gateway)

        # register an unsubscribe for SIGTERM
        def cancel_request(unused_signum, unused_frame):
            log.debug(f"Canceling stream")
            stream.cancel()

        signal.signal(signal.SIGTERM, cancel_request)

        for inp in stream:
            log.debug(f"Got message {inp}")
            start_time = time.time()
            # Add the request_id and timeout to the context
            inp.context["_bw_request_id"] = inp.request_id
            inp.context["_bw_request_timeout"] = inp.timeout.ToJsonString()

            if inp.context.get("_bw_message_type") == "pickle":
                pkl = base64.b16decode(inp.payload)
                payload = pickle.loads(pkl)
            else:
                payload = json.loads(inp.payload)
            try:
                fn(swarm, payload, inp.context)
                messages_received.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name, 0
                ).inc()
            except Exception as e:
                messages_received.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name, 1
                ).inc()
                (_, _, tb) = sys.exc_info()
                error_message = traceback.format_exception(type(e), e, tb)
                swarm.respond(
                    f"Exception when handling payload: {error_message}", inp.context
                )
                raise e
            finally:
                end_time = time.time() - start_time
                bee_processing_time.labels(
                    watcher.bee.name, watcher.bee.swarm, self.name
                ).observe(end_time)


# The dict that holds all of the input types
inputs = {"bee": BeeInput, "gateway": GatewayInput}
