from __future__ import annotations

import sys
import os
import logging
import grpc
import pickle
import base64
import json
import structlog

from collections.abc import Callable
from typing import Dict, Union, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

from yaml import safe_load

from google.protobuf.timestamp_pb2 import Timestamp

from .metrics import *
from .proto import swarm_pb2, swarm_pb2_grpc
from .inputs.input_types import BeeInput, GatewayInput, inputs

log = structlog.get_logger()


class DevSwarm:
    def __init__(self):
        self.published = []
        self.responded = []

    def watch(self, fn):
        self.watch_fn = fn

    def input(self, request):
        empty_context = {}
        self.watch_fn(self, request, empty_context)

    def publish(self, object, context):
        self.published.append(object)

    def respond(self, object, context):
        self.responded.append(object)

    def reset(self):
        self.published = []
        self.responded = []


# TODO: These environment variables could be stronger guarantees
# but because we import `Swarm` in swarm/__init__.py
# They are initialized when you run the client and other
# parts of the SDK.
@dataclass_json
@dataclass(repr=True)
class SwarmConfig:
    input: dict
    name: str
    swarm_api_address: str = os.environ.get("BYTEWAX_SWARM_API", "")
    swarm_name: str = os.environ.get("BYTEWAX_SWARM", "")
    token: str = os.environ.get("BYTEWAX_TOKEN", "")
    swarm_version: str = os.environ.get("BYTEWAX_SWARM_VERSION", "1")


class Swarm(object):
    def __init__(self):
        self.config = self.load_config()
        log.debug(f"Swarm config: {self.config}")
        channel = grpc.insecure_channel(self.config.swarm_api_address)

        # Construct the gRPC Bee
        self.bee = swarm_pb2.Bee(
            token=self.config.token,
            name=self.config.name,
            swarm=self.config.swarm_name,
            swarm_version=self.config.swarm_version,
        )

        # Create the gRPC service stub
        self.stub = swarm_pb2_grpc.SwarmServiceStub(channel)

    def load_config(self):
        """Load config from one of multiple locations for easy local dev"""
        if os.path.exists("./bee_conf.yaml"):
            bee_conf = safe_load(open("./bee_conf.yaml", "r"))
        elif os.path.exists("~/bee_conf.yaml"):
            bee_conf = safe_load(open("~/bee_conf.yaml", "r"))
        elif os.path.exists("/etc/bytewax/bee_conf.yaml"):
            bee_conf = safe_load(open("/etc/bytewax/bee_conf.yaml", "r"))
        else:
            raise Exception("Bee config file not found")

        return SwarmConfig.from_dict(bee_conf)

    def watch(
        self,
        fn: Callable[[Union[Swarm, DevSwarm], Union[str, Dict], Dict[Any, Any]], Any],
    ) -> None:
        """Construct the appropriate ~bytewax.swarm.input_types.BeeInput` type and
        call ``watch`` on it.

        This function is called in the context of the runner to pass input
        from the gRPC channel to the registered function in the Bee.

        In order to do so, we need to construct the correct Input type
        for this Bee.

        Args:
            fn (function):
                A function that accepts three arguments: A Swarm
                or DevSwarm, a payload in the form of a `str` or a Dictionary,
                and a `context` dict.
        """

        # Find the input configuration for this bee
        input_bee = self.config.input

        input_bee["swarm_name"] = self.config.swarm_name
        input_bee["user"] = self.config.token
        input_bee["swarm_api_address"] = self.config.swarm_api_address
        input_bee["token"] = self.config.token
        input_bee["swarm_version"] = self.config.swarm_version
        input_bee[
            "queue_name"
        ] = f"{self.config.swarm_name}-{self.config.name}-{self.config.swarm_version}"
        # Construct the input for this bee, defaulting to BeeInput
        input_type = input_bee.get("type", "bee")
        bee_input = inputs[input_type].from_dict(input_bee)
        bee_input.watch(self, fn, watcher=self)

    def publish(self, message, context):
        """Publish `message` to the swarm by sending it to the swarm-api.

        Args:
            message (string): The body of the request to send to the swarm.
            context (dict): The context dictionary for the payload.
        """
        try:
            payload = json.dumps(message)
        except TypeError:
            log.debug("Non-jsonable python object, pickling")
            context["_bw_message_type"] = "pickle"
            pickled = pickle.dumps(message)
            payload = base64.b16encode(pickled).decode("utf-8")

        timeout = Timestamp()
        timeout.FromJsonString(context["_bw_request_timeout"])

        log.info(f"Publishing payload {payload}", request_id=context["_bw_request_id"])

        swarm_message = swarm_pb2.SwarmMessage(
            payload=payload,
            context=context,
            request_id=context["_bw_request_id"],
            timeout=timeout,
        )
        out = swarm_pb2.Out(
            bee=self.bee,
            message=swarm_message,
            request_id=context["_bw_request_id"],
        )
        self.stub.Output(out)
        messages_published.labels(self.bee.name, self.config.swarm_name).inc()

    def respond(self, payload, context):
        """
        Publish a response to the swarm-api to be returned to the caller
        """
        try:
            payload = json.dumps(payload)
        except TypeError:
            log.debug("Non-jsonable python object, pickling")
            context["_bw_message_type"] = "pickle"
            pickled = pickle.dumps(payload)
            payload = base64.b16encode(pickled).decode("utf-8")

        log.info(
            f"Responding with payload {payload}", request_id=context["_bw_request_id"]
        )

        swarm_message = swarm_pb2.SwarmMessage(
            payload=payload,
            context=context,
            request_id=context["request_id"],
        )
        self.stub.Respond(swarm_message, timeout=10.0)
        messages_responded.labels(self.bee.name, self.config.swarm_name).inc()


# Module level storage for functions called during
# swarm.watch from runner.py
bee_func = {}


def register_bee(name: str):
    """Decorator to register a swarm function for a bee

    This function stores functions from client code in
    `bee_func`, for execution during `Swarm.watch`.
    """

    def inner_decorator(
        fn: Callable[[Union[Swarm, DevSwarm], Union[str, Dict], Dict[Any, Any]], Any]
    ):
        bee_func[name] = fn
        return fn

    return inner_decorator
